#!/usr/bin/python
#
# Copyright 2007 Google Inc.
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: diamond@google.com (Stephen Shirley)
#         marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import signal
import subprocess
import sys
import gobject

from tacix import TXXauth
from tacix.TXDBus import *

class NXAgent:
    def __init__( self, nxsession, options ):
        self.options = options
        self.txpath = nxsession.txpath
        self.txsession = nxsession.txsession
        self.logger = self.txpath.get_logger("nxagent")
        self.process = None
        self.counter = None

    # Utilities methods
    def send_signal(self, num):
        try:
            self.logger.info('Send signal %d to agent %d' % (num, self.process.pid))
            os.kill(self.process.pid, num)
            return True
        except OSError, e:
            return False

    def watch_agent_state_watchdog(self, state):
        if state == 'waiting_watchdog':
            self.kill_watchdog()
        elif state == 'suspended':
            self.watchdog_watcher.remove()
            self.watchdog_watcher = None
            self._restore_session()
        else:
            self.logger.error('Something strange happened: State is %s' % state)

    def kill_watchdog(self):
        if self.txsession.GetState() != 'waiting_watchdog':
            return False
        try:
            os.kill(int(self.txsession.GetProperty('watchdog_pid')), signal.SIGTERM)
        except OSError, (errno, strerror):
            self.logger.warn("matched info kill_watchdog, "
                        "got error from kill[%d]: %s\n" % (errno, strerror))
        except Exception, e:
            self.logger.error(e)
        else:
            self.logger.info("matched info kill_watchdog, sent TERM.\n")
        return False


    # Options manipulations functions
    def resume(self, options):
        self.options = options
        self.logger.info("Rewriting the options file")
        self.write_options()

        state = self.txsession.GetState()
        # TODO We must test which state is resumable
        if state not in ('suspended', 'running'):
            return

        if state == 'running':
            # We need to suspend before restore
            self.watchdog_watcher = self.txsession.connect_to_signal("StateChanged", self.watch_agent_state_watchdog, dbus_interface=INTERFACE_SESSION)
            self.logger.debug('Added watchdog watcher')
            self.send_signal(signal.SIGHUP)
            return
        self._restore_session()

    def _restore_session(self):
        if self.txsession.GetState() != 'suspended':
            gobject.timeout_add(300, self._restore_session)
            if self.counter is None:
                self.counter = 100
            self.counter = self.counter -1
            self.logger.debug('Trying to resume for the %s time' % self.counter)
            if self.counter == 0:
                self.counter = None
                return False
            return True
        self.send_signal(signal.SIGHUP)
        self.counter = None
        return False


    def write_options(self):
        opts = ['nx/nx']

        for opt_name in ("keyboard", "id", "clipboard", "geometry",
                "option", "shadow", "shadowmode", "streaming", "defer", "tile",
                "R", "render", "fullscreen", "shpix", "shmem", "link",
                "media", "cups", "samba", "cache", "images",
                "resize", "backingstore", "menu", "shadowuid", "clients",
                "client", "copysize", "cookie", "cleanup" ):
            if opt_name in self.options:
                opts.append('%s=%s' % (opt_name, self.options[opt_name]))
        opts.append("type=" + self.options['type'].replace('unix-', ''))

        open( self.txpath.get_filepath( "options" ), "w").write("%s:%s\n" % (",".join(opts), self.options['display']))


    # Action methods
    def terminate(self):
        self.send_signal(signal.SIGTERM)

    def run(self):
        self.logger.info("Starting session %s\n" % self.options['id'])
        self.logger.debug("Exporting DISPLAY for nxagent:")
        display_tmp = 'nx/nx,options=%s:%s' % (self.txpath.get_filepath( "options" ), self.options['display'])
        os.environ['DISPLAY'] = display_tmp

        if not self.setup_xauth():
            self.txsession.SetState('denied')
            return False
        self.start_agent()
        os.environ['DISPLAY'] = ':%s' % self.options['display']
        return True

    def start_agent(self):
        self.logger.info("Starting nxagent\n")
        self.write_options()
        cmd = []
        cmd.append('/usr/bin/nxagent')
        if self.options['type'] == 'shadow':
            cmd.append('-S')
            cmd.append('-shadow')
            #cmd.append('%(shadowhost)s:%(shadowdisplay)s' % self.options)
            cmd.append(':%(shadowdisplay)s' % self.options)
            cmd.append('-shadowmode')
            cmd.append('1') #Interactive showdowing [1|0]
        else:
            if self.options['rootless'] == '1': cmd.append('-R')
            else: cmd.append('-D')

            if self.options['persistent'] == '1': cmd.append('-persistent')
            else: cmd.append('-nopersistent')

        cmd.append('-name')
        cmd.append('"TaciX - %s@%s:%s - %s" ' % \
                     (self.options['user'], os.uname()[1], self.options['display'], self.options['session']))
        cmd.append(':%s' % self.options['display'])

        self.logger.debug("Agent cmd: %s\n" % cmd)

        self.process = subprocess.Popen(cmd,
                stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self.logger.info("Spawned nxagent pid %d\n" % self.process.pid)

    def setup_xauth(self):
        """Setup Xauthority file using session cookie."""
        authority_file = self.txpath.get_filepath('authority')
        os.environ['XAUTHORITY'] = authority_file
        self.txsession.SetProperty('authority_file', authority_file)

        if not TXXauth.add_cookie_to_xauthority(authority_file, 'localhost:%(display)s' % self.options, self.options['cookie']):
            return False
        if not TXXauth.add_cookie_to_xauthority(authority_file, ':%(display)s' % self.options, self.options['cookie']):
            return False

        if self.options['type'] == 'shadow':
            os.environ['SHADOW_XAUTHORITY'] = self.txpath.get_filepath('authority')
            # Ask authorization
            if not self.txsession.AskShadowPermission():
                return False
        return True


