#!/usr/bin/python
#
# Copyright 2009 Marcelo Boveto Shima.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA    02110-1301, USA.
#
# Author: marceloshima@gmail.com (Marcelo Boveto Shima)

import os
import os.path
import signal
import subprocess
import threading
import traceback
import gobject
import shutil
gobject.threads_init()

from tacix import TXXauth

from tacix.nx.NXAgent import NXAgent
from tacix.nx.NXAgentMonitor import NXAgentMonitor
from tacix.resources.TXResources import TXResources
from tacix.TXDBus import *
from tacix.TXUtils import TXPath

class NXSession:
    def __init__ ( self, sessionid ):
        from dbus.mainloop.glib import DBusGMainLoop
        DBusGMainLoop(set_as_default=True)
        self.loop = gobject.MainLoop()

        signal.signal(signal.SIGHUP, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)

        self.txsession = get_txsession(sessionid)
        self.txpath = TXPath(sessionid, create=True)
        self.logger = self.txpath.get_logger("nxsession")

        self.nxagent = NXAgent(self, self.txsession.GetProperties())
        self.nxagentmonitor = NXAgentMonitor(self)
        self.app_process = None
        self.app_signal = None
        self.app_thread = None

        self.txresources = TXResources(sessionid)

    # Watchers
    def signal_handler(self, num, stack):
        self.logger.info('Received signal %d' % num)
        if self.nxagent != None:
            self.logger.info('Sending signal to agent')
            self.nxagent.send_signal(num)

    def resume(self):
        self.nxagent.resume(self.txsession.GetProperties())

    # Agent actions
    def run( self ):
        os.environ['XDG_SESSION_COOKIE'] = self.txsession.OpenSession("")
        os.environ['NX_CLIENT'] = '/usr/NX/bin/nxclient'

        self.logger.debug("TXManager session openned\n")

        # Add watchers for the sessionapp
        self.txresources.setup()
        if self.nxagent.options['type'] != 'shadow':
            self.logger.debug('Adding watcher for app')
            self.app_signal = self.add_dbus_signal(self.watch_agent_state_apps)
            self.resume_signal = self.add_dbus_signal(self.resume, signal="ResumeSolicited")

        # Run the agent
        if not self.nxagent.run():
            return False
        self.nxagentmonitor.run()
        return True

    def wait( self ):
        self.loop.run()
        if self.app_thread is not None:
            self.app_thread.join()

    # App actions
    def watch_agent_state_apps(self, state):
        self.logger.info("App watcher state is %s" % state)
        if self.app_process is None and state == 'running':
            self.run_app()
        elif state == 'waiting_watchdog':
            # Add a timeout of 5s to kill the watchdog
            gobject.timeout_add(5000, self.kill_watchdog)
        elif state == 'terminated':
            # Add a timeout of 2s to cleanup
            gobject.timeout_add(2000, self.cleanup)

    def kill_watchdog(self):
        if self.txsession.GetState() != 'waiting_watchdog':
            return False
        try:
            os.kill(int(self.txsession.GetProperty('watchdog_pid')), signal.SIGTERM)
        except OSError, (errno, strerror):
            self.logger.warn("matched info kill_watchdog, "
                        "got error from kill[%d]: %s\n" % (errno, strerror))
        except Exception, e:
            self.logger.error(e)
        else:
            self.logger.info("matched info kill_watchdog, sent TERM.\n")
        return False


    def run_rdesktop(self):
        self.logger.info("NXSessionApp ready, launching rdesktop")
        log_file = open(self.txpath.get_filepath('logs/app.log'), 'w')

        if self.nxagent.options['agent_domain'] != '':
            self.nxagent.options['agent_domain'] = '-d ' + self.nxagent.options['agent_domain']
        if self.nxagent.options['application'] != '':
            self.nxagent.options['application'] = '-s ' + self.nxagent.options['application']
        cmd = 'rdesktop -u "%(agent_user)s" -k "%(agent_keyboard)s" %(agent_domain)s %(application)s -f %(agent_server)s' % self.nxagent.options
        process = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE, stdout=log_file, stderr=log_file, close_fds=True)
        process.stdin.write(self.nxagent.options['agent_password'])
        # TODO Verify if the rdesktop is not closed
        process.stdin.close()
        return process

    def run_vncviewer(self):
        self.logger.info("NXSessionApp ready, launching vncviewer")
        log_file = open(self.txpath.get_filepath('logs/app.log'), 'w')

        cmd = 'vnc' % self.nxagent.options
        process = subprocess.Popen(cmd, shell=True, stdout=log_file, stderr=log_file, close_fds=True)
        return process

    def run_xsession(self):
        """Start the session application (startkde/gnome-session/etc)."""
        cmd = '/etc/tacix/Xsession'
        self.logger.info("NXSessionApp ready, launching app")
        log_file = open(self.txpath.get_filepath('logs/app.log'), 'w')

        os.environ['NX_SESSION_TYPE'] = self.nxagent.options['type']
        if 'application' in self.nxagent.options:
            os.environ['NX_APPLICATION'] = self.nxagent.options['application']
        return subprocess.Popen(cmd, stdout=log_file, stderr=log_file, close_fds=True)

    def run_app(self):
        type = self.nxagent.options['type']
        if type == 'windows':
            self.app_process = self.run_rdesktop()
        elif type == 'vnc':
            self.app_process = self.run_vncviewer()
        else:
            self.app_process = self.run_xsession()
        self.app_thread = threading.Thread(target=self.wait_app_process, name='app_thread')
        self.app_thread.start()

    def wait_app_process( self ):
        self.logger.debug("Waiting for nxapp")
        self.app_process.wait()
        self.nxagent.terminate()
        self.logger.debug("nxapp closed")

    def add_dbus_signal(self, callback, signal="StateChanged"):
        try:
            watcher = self.txsession.connect_to_signal(signal, callback, dbus_interface=INTERFACE_SESSION)
            self.logger.debug('Watcher added')
            return watcher

        except Exception, e:
            self.logger.error("Got exception, cleaning up\n")
            trace = traceback.format_exc()
            for line in trace.split('\n'):
                self.logger.error('%s' % line)

        self.logger.debug('Watcher failed to be added')
        return None

    def cleanup_xauth(self):
        """Setup Xauthority file using session cookie."""
        authority_file = self.txpath.get_filepath('authority')
        TXXauth.remove_display_from_xauthority(authority_file, 'localhost:%(display)s' % self.nxagent.options)
        TXXauth.remove_display_from_xauthority(authority_file, ':%(display)s' % self.nxagent.options)

    def cleanup(self, status=None, keep_log=True):
        """Cleanup after various things (X11 locks etc)."""

        self.logger.info('Cleaning up session\n')

        for file in ('/tmp/.X%(display)s-lock' % self.nxagent.options,
                  '/tmp/.nX%(display)s-lock' % self.nxagent.options,
                  '/tmp/.X11-unix/X%(display)s' % self.nxagent.options):
            if os.path.exists(file):
                try:
                    os.remove(file)
                except OSError, e:
                    self.logger.error("Error removing file. %s\n" % e)

        #self.cleanup_xauth()

        # Move if failed or terminated if defined
        self.logger.info('NXSession Exited')
        if not keep_log:
            shutil.rmtree(self.txpath.get_dir())
        elif status is None:
            shutil.move(self.txpath.get_dir(), self.txpath.get_unknown_dir())
        elif status == 'terminated':
            shutil.move(self.txpath.get_dir(), self.txpath.get_terminated_dir())
        elif status == 'failed':
            shutil.move(self.txpath.get_dir(), self.txpath.get_failed_dir())
        self.loop.quit()
        return False

